/* -*-c++-*- Producer - Copyright (C) 2001-2006  Don Burns
 *
 * This library is open source and may be redistributed and/or modified under
 * the terms of the OpenSceneGraph Public License (OSGPL) version 0.0 or
 * (at your option) any later version.  The full license is in LICENSE file
 * included with this distribution, and on the openscenegraph.org website.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * OpenSceneGraph Public License for more details.
 */
#ifndef PRODUCER_OSG_SCENE_HANDLER_DEF
#define PRODUCER_OSG_SCENE_HANDLER_DEF

#include <Producer/Export>

#include <Producer/Camera>
#include <Producer/Timer>

#include <osg/StateSet>
#include <osg/Light>
#include <osg/PolygonMode>
#include <osg/Viewport>
#include <osg/Group>
#include <osg/Switch>
#include <osg/NodeVisitor>
#include <osg/CollectOccludersVisitor>
#include <osg/ArgumentParser>

#include <osgUtil/UpdateVisitor>
#include <osgUtil/CullVisitor>
#include <osgUtil/StateGraph>
#include <osgUtil/RenderStage>


namespace Producer {

class PR_EXPORT OsgSceneHandler : public Camera::SceneHandler, public osg::CullSettings
{
    public:
        enum Options
        {
            NO_SCENEVIEW_LIGHT = 0x0,
            HEADLIGHT = 0x1,
            SKY_LIGHT = 0x2,
            COMPILE_GLOBJECTS_AT_INIT = 0x4,
            AUTOMATIC_CONTEXT_ID = 0x8,
            STANDARD_SETTINGS = HEADLIGHT | COMPILE_GLOBJECTS_AT_INIT 
                | AUTOMATIC_CONTEXT_ID
        };

        OsgSceneHandler( unsigned int options=STANDARD_SETTINGS );
        OsgSceneHandler( int argc, char **argv, unsigned int options=STANDARD_SETTINGS );
        OsgSceneHandler( osg::ArgumentParser &arguments, unsigned int options=STANDARD_SETTINGS );

        virtual bool useAutoView() { return true; }
        //virtual void clear(Camera &);
        virtual void update();
        virtual void cull( Camera &);
        virtual void draw( Camera &);

        virtual void *getScene();
        virtual bool setScene( void *scene, bool replace=true );
        virtual bool loadScene( const std::string &, bool replace=true );

        virtual bool getSceneBounds( Producer::Vec3 &center, double &radius );

        osg::Node *getRootNode();
        osg::State *getState();

        void flushDeletedGLObjects(double& availableTime);

        osg::Viewport *getViewport();
        bool setViewport( int x, int y, unsigned int width, unsigned int height );

        void toggleLighting();
        void toggleWireframe();
        void toggleTexturing();
        void toggleBackfaceCulling();

        void snapshot( char *fileName );

        osgUtil::UpdateVisitor *getUpdateVisitor();
        osgUtil::CullVisitor *getCullVisitor();

    private:

        // We create our own RenderStage to override drawImplementation with one that
        // does not do actions more appropriate to the camera
        class RenderStage: public osgUtil::RenderStage
        {
            public:
#if defined(OSG_RENDERINFO)
                virtual void drawImplementation(osg::RenderInfo& ,osgUtil::RenderLeaf*& );
#else
                virtual void drawImplementation(osg::State& ,osgUtil::RenderLeaf*& );
#endif
        };

        bool                                        _initCalled;

        Options                                     _lightingMode;

        osg::ref_ptr<osg::Group>                    _rootNode;
        osg::ref_ptr<osg::Node>                     _scene;
        osg::ref_ptr<osg::Switch>                   _scribeSwitch;

#if defined(OSG_RENDERINFO)
        osg::RenderInfo                             _renderInfo;
#else
        osg::ref_ptr<osg::State>                    _state;
#endif
        osg::ref_ptr<osg::StateSet>                 _globalStateSet;
        osg::ref_ptr<osg::StateSet>                 _localStateSet;

        osg::ref_ptr<osg::Light>                    _light;

        osg::ref_ptr<osg::NodeVisitor>              _initVisitor;
        osg::ref_ptr<osg::NodeVisitor>              _updateVisitor;

        osg::ref_ptr<osgUtil::CullVisitor>          _cullVisitor;
        osg::ref_ptr<osgUtil::StateGraph>           _rendergraph;
        osg::ref_ptr<OsgSceneHandler::RenderStage>             _renderStage;

        osg::ref_ptr<osg::CollectOccludersVisitor>  _collectOccludersVisitor;

        const Producer::Timer *_timer;
        Producer::Timer_t _t0;
        osg::ref_ptr<osg::FrameStamp>               _frameStamp;

        bool                                        _requiresFlush;

        bool _lighting;
        bool _texturing;
        bool _backfaceCulling;
        osg::PolygonMode::Mode _polygonMode;
        bool _scribe;

        void _constructorImplementation( osg::ArgumentParser &arguments, unsigned int options );
        void _init(void);
        void _setDefaults( unsigned int options );
        void _initScribeSwitch(void);

        void _cullStage(
            const osg::Matrixd& projection,
            const osg::Matrixd& modelview,
            osgUtil::CullVisitor* cullVisitor, 
            osgUtil::StateGraph* rendergraph, 
            OsgSceneHandler::RenderStage *renderStage );


        // The following added because of the scattering of Stereo functionality
        // throughout osgUtil.  A cleaner approach would be to be completely 
        // Stereo unaware here and simply allow the stereo functionality to take
        // place in the Camera.  This would require a great deal of rewrite of
        // osgUtil, so until we rewrite more of osgUtil functionality, this is
        // the short cut.  
        //                              DB
        virtual void _cullMono( Camera &);
        virtual void _cullStereo( Camera &);
        virtual void _drawMono( Camera &);
        virtual void _drawStereo( Camera &);

        osg::ref_ptr<osgUtil::CullVisitor>          _cullVisitorLeft;
        osg::ref_ptr<osgUtil::CullVisitor>          _cullVisitorRight;
        osg::ref_ptr<osgUtil::StateGraph>           _rendergraphLeft;
        osg::ref_ptr<osgUtil::StateGraph>           _rendergraphRight;
        osg::ref_ptr<OsgSceneHandler::RenderStage>             _renderStageLeft;
        osg::ref_ptr<OsgSceneHandler::RenderStage>             _renderStageRight;

        unsigned long _stereoFrame;

        static int _contextIDCounter;
};

}

#endif
