/* -*-c++-*- Producer - Copyright (C) 2001-2004  Don Burns
 *
 * This library is open source and may be redistributed and/or modified under
 * the terms of the OpenSceneGraph Public License (OSGPL) version 0.0 or
 * (at your option) any later version.  The full license is in LICENSE file
 * included with this distribution, and on the openscenegraph.org website.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * OpenSceneGraph Public License for more details.
 */
#include <iostream>
#include <Producer/Export>
#include <Producer/RenderSurface>

#ifdef _OSX_AGL_IMPLEMENTATION

#include <Carbon/Carbon.h>
#include <OpenGL/OpenGL.h>
#include <AGL/agl.h>

using namespace std;
using namespace Producer;

static OpenThreads::Mutex s_init_mutex;

// get a double value from a dictionary
static double getDictDouble (CFDictionaryRef refDict, CFStringRef key)
{
   double double_value;
   CFNumberRef number_value = (CFNumberRef) CFDictionaryGetValue(refDict, key);
   if (!number_value) // if can't get a number for the dictionary
       return -1;  // fail
   if (!CFNumberGetValue(number_value, kCFNumberDoubleType, &double_value)) // or if cant convert it
        return -1; // fail
    return double_value; // otherwise return the long value
}

bool RenderSurface::realize(Producer::VisualChooser *vc, GLContext shared_context)
{
    // we are already realized, so simply return
    
    if (_realized) return _realized;
        

    if( shared_context != 0L )
        _sharedGLContext = shared_context;
    else
        _sharedGLContext = _globallySharedGLContext;

    if ( _init() == false )
        return false;
    
    _realized = true;
    _realizeBlock->release();

    std::vector <Producer::ref_ptr<Callback> >::iterator p;
    for( p = _realizeCallbacks.begin(); p != _realizeCallbacks.end(); p++ )
    {
        if( (*p).valid() )
            (*(*p).get())( *this );
    }
    _realizeCallbacks.clear();

    return _realized;

}

bool RenderSurface::makeCurrent(bool) 
{
    //s_init_mutex.lock();
    aglSetCurrentContext(_glcontext);
    //s_init_mutex.unlock();
    return true;
}

void RenderSurface::swapBuffers()
{
    //s_init_mutex.lock();
    aglSwapBuffers(_glcontext);
    //s_init_mutex.unlock();
}

void RenderSurface::run()
{
    _init();

    if( _threadReady != 0L)
        _threadReady->block();
}

void RenderSurface::sync(int)
{
}

void RenderSurface::fullScreen( bool flag )
{
    if (flag == _isFullScreen) 
        return;
    
    if (!_realized) {
        _isFullScreen = flag;
        return;
    }
    
    s_init_mutex.lock();
    
    aglSetCurrentContext (_glcontext);
    aglSetDrawable(_glcontext, NULL);
    aglSetFullScreen(NULL, 0,0,0,0);
    _releaseDisplayOrWindow();
    
    _isFullScreen = flag;
    
    _captureDisplayOrWindow();
    
    // bind context
    if (_isFullScreen) {
        _glcontext = _fullscreenGlcontext;
        _visualInfo = _fullscreenVisualInfo;
        aglSetCurrentContext (_glcontext);
        aglEnable (_glcontext, AGL_FS_CAPTURE_SINGLE);
        if (!aglSetFullScreen (_glcontext, 0, 0, 0, 0)) {
            std::cerr << "Rendersurface: could not switch to fullscreen" << std::endl;
        }
    } else {
        _glcontext = _windowGlcontext;
        _visualInfo = _windowVisualInfo;
        aglSetCurrentContext (_glcontext);
        aglSetDrawable(_glcontext, GetWindowPort(_win));
        ShowWindow(_win);
    }
    
    s_init_mutex.unlock();

}

unsigned int RenderSurface::getNumberOfScreens()
{
    if( _numScreens == UnknownAmount )
    {
        if( CGGetActiveDisplayList( 0, NULL, &_numScreens ) != CGDisplayNoErr )
        {
            std::cerr << "CGGetActiveDisplayList() falied\n";
            _numScreens = UnknownAmount;
            return 0;
        }
    }
//    _numScreens = 1;
    return _numScreens;
}


static Display *_openDisplay(unsigned int num)
{
    static Display *displays = NULL;
    static CGDisplayCount numDisplays = 0;

    if( displays == NULL )
    {
        if( CGGetActiveDisplayList( 0, NULL, &numDisplays ) != CGDisplayNoErr )
        {
            std::cerr << "CGGetActiveDisplayList() falied\n";
            return NULL;
        }

        if( numDisplays == 0 )
        {
            std::cerr << "openDisplay: No Displays!\n";
            return NULL;
        }

        displays = new Display[numDisplays];
        if( CGGetActiveDisplayList( numDisplays, displays, &numDisplays ) != CGDisplayNoErr )
        {
            std::cerr << "CGGetActiveDisplayList() falied\n";
            return NULL;
        }
    }

    if( num > numDisplays )
    {
        std::cerr << "OpenDisplay(): " << num << " is out of range.\n";
        return NULL;
    }

    return &displays[num];
}

static pascal OSStatus RenderSurfaceWindowEventHandler(EventHandlerCallRef myHandler, EventRef event, void* userData) 
{
    WindowRef           window;
    Rect                bounds;
    UInt32              whatHappened;
    OSStatus            result = eventNotHandledErr; /* report failure by default */
    

    RenderSurface* rs = (RenderSurface*)userData;
    if (!rs)
        return result;
    
    GetEventParameter(event, kEventParamDirectObject, typeWindowRef, NULL,
                         sizeof(window), NULL, &window);
 
    whatHappened = GetEventKind(event);
    
    
    switch (whatHappened)
        {
            case kEventWindowBoundsChanging:
                // left the code in, but it is not used, because of window-refreshing issues...
                GetEventParameter( event, kEventParamCurrentBounds, typeQDRectangle, NULL, sizeof(Rect), NULL, &bounds );
                rs->setWindowRectangle(bounds.left, bounds.top, bounds.right - bounds.left, bounds.bottom - bounds.top, true);
                result = noErr;
                break;
                    
            case kEventWindowBoundsChanged:
                InvalWindowRect(window, GetWindowPortBounds(window, &bounds));
                
                rs->setWindowRectangle(bounds.left, bounds.top, bounds.right - bounds.left, bounds.bottom - bounds.top, true);
                result = noErr;
                break;
 
            default: 
                /* If nobody handled the event, it gets propagated to the */
                /* application-level handler. */
                break;
 
        }
 
    return result;
}



bool RenderSurface::_captureDisplayOrWindow()
{
    
    _dpy = _openDisplay(_screen);    
    
    if (_isFullScreen)
    {
        
        CGDisplayCapture(*_dpy);
        
        if( _screenWidth == UnknownDimension )
            _screenWidth = CGDisplayPixelsWide(*_dpy);

        if( _screenHeight == UnknownDimension )
            _screenHeight = CGDisplayPixelsHigh(*_dpy);

        
        // add next line and on following line replace hard coded depth and refresh rate
        CGRefreshRate refresh =  getDictDouble (CGDisplayCurrentMode (*_dpy), kCGDisplayRefreshRate);  
        CFDictionaryRef display_mode_values =
            CGDisplayBestModeForParametersAndRefreshRate(*_dpy, CGDisplayBitsPerPixel (*_dpy), 
                            _screenWidth,_screenHeight,  
                            refresh,  
                            NULL);

        // save current display settings
        _oldDisplayMode = CGDisplayCurrentMode(*_dpy);  
                                  
        CGDisplaySwitchToMode(*_dpy, display_mode_values);                             

    }
    else
    {
    
        if( _windowWidth == UnknownDimension )
            _windowWidth = (unsigned int)((float)CGDisplayPixelsWide(*_dpy) * 1.0f);

        if( _windowHeight == UnknownDimension )
            _windowHeight = (unsigned int)((float)CGDisplayPixelsHigh(*_dpy) * 1.0f);
    
        // create window
        if (_win == NULL) {
            Rect bounds = {_windowY, _windowX, _windowY + _windowHeight, _windowX + _windowWidth};

            OSStatus err = 0;
            if (usesBorder())
            {
                err = CreateNewWindow(kDocumentWindowClass,  kWindowLiveResizeAttribute | kWindowStandardDocumentAttributes | kWindowStandardHandlerAttribute, &bounds, &_win);
            } else {
                cerr << "Producer creating borderless window (X:" << _windowX << ",  Y:" << _windowY << ",  width:" << _windowWidth << ",  height:" << _windowHeight << ")" << endl;
                err = CreateNewWindow(kDocumentWindowClass,  kWindowNoTitleBarAttribute | kWindowNoShadowAttribute | kWindowResizableAttribute | kWindowLiveResizeAttribute, &bounds, &_win);
            }
            if (err != 0) {
                std::cerr << "Producer: could not create window: " << err << std::endl;
                _win = NULL;
                return false;
            }

            
            // install resize event handler
            EventTypeSpec   windEventList[] = {{ kEventClassWindow, kEventWindowBoundsChanged}};
            InstallWindowEventHandler(_win, NewEventHandlerUPP(RenderSurfaceWindowEventHandler),  1, windEventList, this, NULL);

        }
    }
    return true;
}

void RenderSurface::_releaseDisplayOrWindow() {

    if (_isFullScreen) {
        if (_dpy) {
            CGDisplaySwitchToMode(*_dpy, _oldDisplayMode);
            CGDisplayRelease(*_dpy);
        }
    } else
        HideWindow(_win);
}

bool RenderSurface::_init()
{
    s_init_mutex.lock();
    
    // capture display / window
    if (!_captureDisplayOrWindow())
        return false;
 
    // Set up the VisualChooser 
    VisualChooser *vc; 
    if(_visualChooser.valid()) vc = _visualChooser.get();
    else
    {
      vc = new VisualChooser;
      vc->setSimpleConfiguration(_drawableType != DrawableType_PBuffer);
    }
       
    // Create context for window
    VisualInfo *vinfo = vc->choose(_dpy, 0);
    if(vinfo)
    {
      _windowGlcontext = aglCreateContext (*vinfo, _sharedGLContext);
      _windowVisualInfo = vinfo;
    }

    // Create context for fullscreen
    vinfo = vc->choose(_dpy, 1);
    if(vinfo)
    {
      _fullscreenGlcontext = aglCreateContext (*vinfo, _windowGlcontext);
      _fullscreenVisualInfo = vinfo;
    }

    // Initialize globallySharedGLContext if global sharing is requested and 
    // the global variable is still NULL.
    // Caveat is if the user attempted to share GL contexts already through realize()
    if( _shareAllGLContexts == true && 
            _globallySharedGLContext == 0L &&
            _sharedGLContext == 0L )
    {
        _globallySharedGLContext = _sharedGLContext = _windowGlcontext;
    }

    // attaching context to window/display
    if (_isFullScreen) {
        _glcontext = _fullscreenGlcontext;
        _visualInfo = _fullscreenVisualInfo;
        makeCurrent();
        aglSetFullScreen(_glcontext, 0, 0, 0, 0);
    } else {
        _glcontext = _windowGlcontext;
        _visualInfo = _windowVisualInfo;
        makeCurrent();
        aglSetDrawable(_glcontext, GetWindowPort(_win));
        ShowWindow(_win);
    }

    // Add sync support by default:
    GLint swap = 1;
    aglSetInteger (_glcontext, AGL_SWAP_INTERVAL, &swap);
    
    s_init_mutex.unlock();
    
    return true;
}

void RenderSurface::_fini()
{
    if (_realized) {
        
        s_init_mutex.lock();

               
        aglSetDrawable(_glcontext, NULL);
        aglSetCurrentContext(NULL);
        aglDestroyContext(_windowGlcontext);
        aglDestroyContext(_fullscreenGlcontext);    
        if (_fullscreenVisualInfo) aglDestroyPixelFormat( *_fullscreenVisualInfo ); 
        if (_windowVisualInfo) aglDestroyPixelFormat( *_windowVisualInfo ); 
        
        _releaseDisplayOrWindow();
        
        if (_win) DisposeWindow(_win);
                                            
        
        s_init_mutex.unlock();

    }                                                
                                        
}

OSStatus aglReportError(void)
{
    GLenum err = aglGetError();
    if (AGL_NO_ERROR != err) { 
        char errStr[256];
        sprintf (errStr, "AGL: %s",(char *) aglErrorString(err));
        std::cerr << "Error - " << errStr << endl;
    }
    if (err == AGL_NO_ERROR)
        return noErr;
    else
        return (OSStatus) err;
}

void RenderSurface::_setBorder(bool flag)
{
   OSStatus err = noErr;
   
   if (_isFullScreen) return;

   Rect rectPortOld;
   GetWindowBounds(_win, kWindowStructureRgn, &rectPortOld);
   
   if (flag)
   {
     // switch on window borders
     if (ChangeWindowAttributes(_win,  kWindowStandardDocumentAttributes | kWindowStandardHandlerAttribute, kWindowNoTitleBarAttribute | kWindowNoShadowAttribute | kWindowResizableAttribute) != noErr)
     {
       cerr << "Change of window attributes failed!" << endl;
     }
     Rect titleRect;
     GetWindowBounds(_win, kWindowTitleBarRgn, &titleRect);

     VisualChooser *vc = getVisualChooser();
     VisualInfo *vinfo = vc->choose(_dpy, 0);
     
     if (_realized) {
        s_init_mutex.lock();
        if (!aglSetDrawable(_glcontext, NULL))
          err = aglReportError();
        if (!aglSetCurrentContext(NULL))
          err = aglReportError();
        if (!aglDestroyContext(_glcontext))
          err = aglReportError();

        AGLContext myglcontext = aglCreateContext(*vinfo, NULL);
        if (!myglcontext)
          err = aglReportError();
        _glcontext = myglcontext;
        _windowGlcontext = myglcontext;

        if (!aglSetDrawable(myglcontext, GetWindowPort(_win)))
          err = aglReportError(); 
        if (!aglSetCurrentContext(myglcontext))
          err = aglReportError();
        if (!aglUpdateContext(myglcontext))
          err = aglReportError();

        glViewport (0, 0, rectPortOld.right - rectPortOld.left, rectPortOld.bottom - rectPortOld.top);
        
        //InvalWindowRect(_win,  &rectPortOld);

        s_init_mutex.unlock();
     }
     
   } else {
     // switch off window borders
     // cerr << "RenderSurface::_setBorder(false) - called!" << endl;
     
     //GetWindowBounds(_win, kWindowTitleBarRgn, &titlePort);
     if (ChangeWindowAttributes(_win,  kWindowNoTitleBarAttribute | kWindowNoShadowAttribute, kWindowStandardDocumentAttributes |  kWindowLiveResizeAttribute) != noErr)
     //if (ChangeWindowAttributes(winref,  kWindowStandardHandlerAttribute, kWindowNoAttributes) != noErr)
     {
       cerr << "Change of window attributes failed!" << endl;
     }

     Rect rectPort;
     GetWindowPortBounds(_win, &rectPort);
     rectPort.bottom += rectPortOld.top;
     rectPort.top = rectPortOld.top;
   
     SetPortBounds(GetWindowPort(_win), &rectPort);
   
     Rect dummyRect;
     GetPortBounds(GetWindowPort(_win), &dummyRect);

     if (SetWindowBounds(_win, kWindowStructureRgn, &rectPort) != noErr)
     {
       cerr << "Window strucrgn resize failed!" << endl;
     }
     if (SetWindowBounds(_win, kWindowContentRgn, &rectPort) != noErr)
     {
       cerr << "Window contenrgn resize failed!" << endl;
     }
     SetPortWindowPort(_win);

     VisualChooser *vc = getVisualChooser();
     VisualInfo *vinfo = vc->choose(_dpy, 0);
     
     if (_realized) {
        s_init_mutex.lock();
        if (!aglSetDrawable(_glcontext, NULL))
          err = aglReportError();
        if (!aglSetCurrentContext(NULL))
          err = aglReportError();
        if (!aglDestroyContext(_glcontext))
          err = aglReportError();

        AGLContext myglcontext = aglCreateContext(*vinfo, NULL);
        if (!myglcontext)
          err = aglReportError();
        _glcontext = myglcontext;
        _windowGlcontext = myglcontext;

        if (!aglSetDrawable(myglcontext, GetWindowPort(_win)))
          err = aglReportError(); 
        if (!aglSetCurrentContext(myglcontext))
          err = aglReportError();
        if (!aglUpdateContext(myglcontext))
          err = aglReportError();

        glViewport (0, 0, rectPort.right - rectPort.left, rectPort.bottom - rectPort.top);

        s_init_mutex.unlock();
     }
   }
}

void RenderSurface::_useCursor(bool flag)
{
    if (_useCursorFlag==flag) return;
    
    _useCursorFlag = flag;

    if( _realized )
    {
        if (_useCursorFlag)
        {
            // switching on cursor
            CGDisplayShowCursor(*_dpy);
        }
        else
        {
            // switching off cursor
            CGDisplayHideCursor(*_dpy);
        }
    }
}

void RenderSurface::_setCursor(Cursor)
{
}

void RenderSurface::_changeCursor(Cursor cursor)
{ 
    _setCursor(cursor); 
}

void RenderSurface::_useOverrideRedirect( bool )
{
    // Ignored on OSX CGL.  Flag value remains false.
}

void RenderSurface::_setCursorToDefault()
{
    _setCursor(_defaultCursor);
}

void RenderSurface::_resizeWindow()
{
    if (_isFullScreen) {
        _windowWidth =  CGDisplayPixelsWide(*_dpy);
        _windowHeight =  CGDisplayPixelsHigh(*_dpy);
    } 
    else 
    {
        Rect bound;
        GetWindowPortBounds (_win, &bound);
        _windowWidth = bound.right - bound.left;
        _windowHeight = bound.bottom - bound.top;
        
        // the window is resized by the system, we need only to udpate the context:
        
        makeCurrent();
        aglUpdateContext(_glcontext);
    }

}

void RenderSurface::_positionPointer(int x, int y)
{
    if( _realized )
    {
        CGPoint point;
        point.x = x;
        point.y = y;
        CGDisplayMoveCursorToPoint(*_dpy, point);
    }
}


void RenderSurface::_computeScreenSize(unsigned int &width, unsigned int &height) const
{
    width =  CGDisplayPixelsWide(*_dpy);
    height =  CGDisplayPixelsHigh(*_dpy);
}


void RenderSurface::bindPBufferToTexture(BufferType /*buffer*/) const
{
    // Ok... a bit of a hack.
    glCopyTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, 0, 0, _windowWidth, _windowHeight);
}


void RenderSurface::_setWindowName( const std::string &name )
{
    if( !_realized )
        return;
}

void RenderSurface::_initThreads()
{
    // Ignored...
}

#endif
